"""Test plotting functions (mostly for pseudovariables)"""
import unittest
import numpy as np
import pyspedas
from pyspedas import themis
from pytplot import store_data, options, timespan, tplot, tplot_options, degap, tplot_names, del_data, ylim
import pytplot

# Set this to false for Github CI tests, set to True for interactive use to see plots.
global_display = False
default_trange=['2007-03-23','2007-03-24']
class PlotTestCases(unittest.TestCase):
    """Test plot functions."""

    def test_markers_and_symbols(self):
        # Regression test for lineplot crash when marker sizes are set
        # Taken from pytplot markers and symbols notebook in pyspedas_examples
        del_data('*')
        store_data('data', data={'x': [1, 2, 3, 4, 5, 6], 'y': [1, 1, 1, 1, 1, 1]})
        timespan('1970-01-01',10,'seconds')
        tplot('data', display=global_display, save_png='simple_lineplot.png')
        options('data', 'marker', 'X')
        tplot('data', display=global_display, save_png='markers_lineplot.png')
        options('data', 'linestyle', 'None')
        tplot('data', display=global_display, save_png='noline_lineplot.png')
        options('data', 'marker_size', 200)
        tplot('data', display=global_display, save_png='markersize_lineplot.png')
        options('data', 'line_style_name', 'solid')
        tplot('data', display=global_display, save_png='markersize_nosymbols_lineplot.png')
        options('data', 'marker_size', 20)
        tplot('data', display=global_display, save_png='markersize20_nosymbols_lineplot.png')
        options('data', 'markevery', 2)
        tplot('data', display=global_display, save_png='markevery_lineplot.png')
        options('data', 'marker', 'H')
        tplot('data', display=global_display, save_png='hexagons_lineplot.png')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests

    def test_markers_and_symbols_error_bars(self):
        # Regression test for lineplot crash when marker sizes are set
        # Taken from pytplot markers and symbols notebook in pyspedas_examples
        del_data('*')
        store_data('data', data={'x': [1, 2, 3, 4, 5, 6], 'y': [1, 1, 1, 1, 1, 1], 'dy':[0.25]*6})
        timespan('1970-01-01',10,'seconds')
        tplot('data', display=global_display, save_png='simple_lineplot_errbars.png')
        options('data', 'marker', 'X')
        tplot('data', display=global_display, save_png='markers_lineplot_errbars.png')
        options('data', 'line_style', 'None')
        tplot('data', display=global_display, save_png='symbols_lineplot_errbars.png')
        options('data', 'marker_size', 200)
        tplot('data', display=global_display, save_png='markersize_lineplot_errbars.png')
        options('data', 'symbols', False)
        tplot('data', display=global_display, save_png='markersize_nosymbols_lineplot_errbars.png')
        options('data', 'marker_size', 20)
        tplot('data', display=global_display, save_png='markersize20_nosymbols_lineplot_errbars.png')
        options('data', 'markevery', 2)
        tplot('data', display=global_display, save_png='markevery_lineplot_errbars.png')
        options('data', 'marker', 'H')
        tplot('data', display=global_display, save_png='hexagons_lineplot_errbars.png')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests

    def test_timebars(self):
        from pytplot import timebar
        del_data('*')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests
        themis.fgm(probe='c', trange=default_trange)
        timebar(t=10000.0,varname='thc_fge_dsl',databar=True,color='black')
        timebar(t=-10000.0,varname='thc_fge_dsl',databar=True,color='red', dash=True)
        timebar(t=pytplot.time_double('2007-03-23/14:00'),varname='thc_fge_btotal',color='magenta')
        timebar(t=pytplot.time_double('2007-03-23/14:30'),color='blue')
        timebar(t=pytplot.time_double('2007-03-23/15:30'),color='green', dash=True)
        tplot_options('title', 'Databars at +/- 10000 top panel, timebars at 14:30 and 15:30 all panels, timebar at 14:00 bottom panel, linestyles and colors as specified')
        tplot(['thc_fge_dsl','thc_fge_btotal'], display=global_display, save_png='timebars.png')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests


    def test_line_pseudovariables(self):
        del_data("*")
        # Test that tplot variables with different number of traces can be combined into a pseudovariable and plotted correctly.
        # Both plots should have 4 properly labeled traces.
        themis.fgm(probe='c', trange=default_trange)
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests
        store_data('comb_3_1',data=['thc_fge_dsl','thc_fge_btotal'])
        store_data('comb_1_3',data=['thc_fge_btotal','thc_fge_dsl'])
        tplot_options('title', 'Pseudovariable with one+three line traces')
        tplot('comb_1_3',save_png='pseudovars_comb_1_3',display=global_display)
        tplot_options('title', 'Pseudovariable with three+one line traces')
        tplot('comb_3_1', save_png='pseudovars_comb_3_1',display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests



    def test_pseudovar_color_options(self):
        del_data("*")
        themis.fgm(probe='c',trange=default_trange)
        timespan('2007-03-23', 1, 'days')
        store_data('test_pseudo_colors',data=['thc_fge_dsl','thc_fge_btotal'])
        # Set the color option on the pseudovariable (4 traces total, so 4 colors)
        options('test_pseudo_colors','color',['k','r','g','b'])
        tplot_options('title', 'Trace colors should be black, red, green, blue')
        tplot('test_pseudo_colors', save_png='test_pseudo_colors_4color', display=global_display ) # should plot without "incorrect number of line colors" messages
        options('test_pseudo_colors','color',['k']) # All black
        tplot_options('title', 'Trace colors all black')
        tplot('test_pseudo_colors', save_png='test_pseudo_colors_allsamecolor', display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_var_line_options(self):
        del_data("*")
        themis.fgm(probe='c',trange=default_trange)
        timespan('2007-03-23', 1, 'days')
        options('thc_fgs_dsl','line_style',['solid','dot','dash'])
        tplot_options('title', 'Line styles solid, dot, dash')
        tplot('thc_fgs_dsl',save_png='test_linestyle_3styles',display=global_display)
        options('thc_fgs_dsl','line_style','dot') # gets used for all lines
        tplot_options('title', 'Line styles all dot')
        tplot('thc_fgs_dsl',save_png='test_linestyle_allsame',display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_ylim(self):
        del_data("*")
        themis.fgm(probe='c',trange=default_trange)
        timespan('2007-03-23', 1, 'days')
        ylim('thc_fgs_dsl',-100, 100)
        tplot_options('title', 'Y limit [-100, 100]')
        tplot('thc_fgs_dsl',save_png='test_ylim',display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_is_pseudovar(self):
        del_data("*")
        from pytplot import is_pseudovariable
        themis.fgm(probe='c',trange=default_trange)
        timespan('2007-03-23', 1, 'days')
        store_data('test_pseudo_colors',data=['thc_fge_dsl','thc_fge_btotal'])
        self.assertTrue(is_pseudovariable('test_pseudo_colors'))
        self.assertFalse(is_pseudovariable('thc_fgs_dsl'))
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_count_traces(self):
        del_data("*")
        from pytplot import count_traces
        themis.fgm(probe='c',trange=default_trange)
        store_data('test_pseudo_colors',data=['thc_fge_dsl','thc_fge_btotal'])
        tr_fge=count_traces('thc_fge_dsl')
        tr_btotal=count_traces('thc_fge_btotal')
        tr_pseudo=count_traces('test_pseudo_colors')
        self.assertEqual(tr_fge,3)
        self.assertEqual(tr_btotal,1)
        self.assertEqual(tr_pseudo,4)
        # We need to ensure thc_fge_dsl has a 'v' component, so we can pretend this variable is a spectrogram
        fgs_data = pytplot.get_data('thc_fge_dsl')
        fgs_meta = pytplot.get_data('thc_fge_dsl', metadata=True)
        store_data('thc_fge_dsl',data={'x':fgs_data.times, 'y':fgs_data.y, 'v':[0, 1, 2]})
        options('thc_fge_dsl','spec',1)
        tr_pseudo_spec=count_traces('test_pseudo_colors')
        self.assertEqual(tr_pseudo_spec,1)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests


    def test_pseudovar_line_options(self):
        del_data("*")
        themis.fgm(probe='c',trange=default_trange)
        timespan('2007-03-23', 1, 'days')
        store_data('test_pseudo_lineopts',data=['thc_fge_dsl','thc_fge_btotal'])
        # Set the line_style on the pseudovariable (4 traces total, so 4 styles)
        options('test_pseudo_lineopts','line_style',['dot','dash','solid','dash_dot'])
        tplot_options('title', 'Pseudovar line styles dot, dash, solid, dash_dot')
        tplot('test_pseudo_lineopts', save_png='test_pseudo_lineopts_4style',display=global_display)
        # Set all traces to the same style
        options('test_pseudo_lineopts','line_style','dot')
        tplot_options('title', 'Pseudovar line styles all dot')
        tplot('test_pseudo_lineopts', save_png='test_pseudo_lineopts_allsame',display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_specplot_optimizations(self):
        del_data("*")
        ask_vars = themis.ask(trange=['2013-11-05', '2013-11-06'])
        timespan('2013-11-05',1,'days')
        # Should plot without errors, show something other than all-blue or vertical lines
        tplot_options('title', 'Should be mostly dark with a few lighter features')
        tplot(['thg_ask_atha'],save_png='thg_ask_atha',display=global_display)
        options('thg_ask_atha','y_no_resample',1)
        tplot_options('title', 'Should be mostly dark with a few lighter features')
        tplot('thg_ask_atha', save_png='thg_ask_atha_no_resample',display=global_display)
        tplot_options('title', 'Should be somewhat lighter now with logarithmic z scale')
        options('thg_ask_atha','zlog',"log")
        tplot('thg_ask_atha', save_png='thg_ask_atha_no_resample_zlog',display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_themis_peef_bins(self):
        del_data("*")
        themis.esa(probe='a',trange=['2016-12-11','2016-12-12'])
        timespan('2016-12-11',1,'days')
        options('tha_peef_en_eflux','yrange', [1000,3000])
        tplot_options('title', 'Logarithmic bin boundaries: should be a boundary at y~=1947 eV, just below the tick mark at y=2000')
        tplot('tha_peef_en_eflux',save_png='tha_peef_en_eflux',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_themis_peef_bins_interp_x(self):
        del_data("*")
        themis.esa(probe='a',trange=['2016-12-11','2016-12-12'])
        timespan('2016-12-11',1,'hours')
        options('tha_peef_en_eflux','yrange', [1000,3000])
        options('tha_peef_en_eflux', 'x_interp',1)
        options('tha_peef_en_eflux', 'x_interp_points', 500)
        tplot_options('title', 'Interpolated along time axis')
        tplot('tha_peef_en_eflux',save_png='tha_peef_en_eflux_interp_x',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests
        options('tha_peef_en_eflux', 'x_interp',0) # reset for other tests


    def test_themis_peef_bins_interp_y(self):
        del_data("*")
        themis.esa(probe='a',trange=['2016-12-11','2016-12-12'])
        timespan('2016-12-11',1,'hours')
        options('tha_peef_en_eflux','yrange', [1000,3000])
        options('tha_peef_en_eflux', 'y_interp',1)
        options('tha_peef_en_eflux', 'y_interp_points', 200)
        tplot_options('title', 'Interpolated along y axis')
        tplot('tha_peef_en_eflux',save_png='tha_peef_en_eflux_interp_y',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests
        options('tha_peef_en_eflux', 'y_interp',0) # reset for other tests

    def test_themis_peef_bins_interp_both(self):
        del_data("*")
        themis.esa(probe='a',trange=['2016-12-11','2016-12-12'])
        timespan('2016-12-11',1,'hours')
        options('tha_peef_en_eflux','yrange', [1000,3000])
        options('tha_peef_en_eflux', 'x_interp',1)
        options('tha_peef_en_eflux', 'x_interp_points', 500)
        options('tha_peef_en_eflux', 'y_interp',1)
        options('tha_peef_en_eflux', 'y_interp_points', 200)
        tplot_options('title', 'Interpolated along both x and y axes')
        tplot('tha_peef_en_eflux',save_png='tha_peef_en_eflux_interp_both',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests
        options('tha_peef_en_eflux', 'x_interp',0) # reset for other tests
        options('tha_peef_en_eflux', 'y_interp',0) # reset for other tests

    def test_mms_epsd_specplot(self):
        del_data("*")
        from pytplot import options
        from pyspedas import mms_load_dsp
        timespan('2015-08-01',1,'days')
        # Logarithmic Y scale with lowest bin boundary = 0.0 by linear extrapolation from bin centers
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype=['epsd', 'bpsd'], level='l2', data_rate='fast')
        # options('mms1_dsp_epsd_omni','yrange',[8.0,130000.0])
        tplot(['mms1_dsp_epsd_omni', 'mms1_dsp_bpsd_omni'], save_png='mms1_epsd_omni', display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_elfin_specplot(self):
        del_data("*")
        import pyspedas
        # ELFIN data with V values that oscillate, the original problem that resulted in the resample, this is an angular distrubtion
        timespan('2021-07-14/11:55',10,'minutes')
        epd_var = pyspedas.projects.elfin.epd(trange=['2021-07-14/11:55', '2021-07-14/12:05'], probe='a', level='l2', type_='nflux', fullspin=False)
        tplot_options('title', 'ELFIN data with time-varying bins, should render accurately')
        tplot('ela_pef_hs_nflux_ch0', display=global_display, save_png='ELFIN_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests

    def test_fast_specplot(self):
        del_data("*")
        import pyspedas
        # FAST TEAMS has fill values -1e31 in V, top is an energy distribution, the bottom two are pitch angle distributions
        teams_vars = pyspedas.projects.fast.teams(['1998-09-05', '1998-09-06'], level='k0')
        timespan('1998-09-05',1,'days')
        tplot_options('title', 'Fill should be removed, bottom two panels should go to Y=-90 deg')
        # Specify variables to plot as a space-delimited string with wildcards
        tplot('*H+ *H+_low *H+_high', display=global_display, save_png='TEAMS_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # reset to avoid interfering with other tests

    def test_themis_esa_specplot(self):
        del_data("*")
        import pyspedas
        # THEMIS ESA has monotonically decreasing energies, time varying energies, and also has fill
        esa_vars = pyspedas.projects.themis.esa(trange=['2016-07-23', '2016-07-24'], probe='a')
        timespan('2016-07-23',1,'days')
        tplot_options('title', 'Decreasing and time-varying energies, fillvals, should render correctly')
        tplot('tha_peef_en_eflux', display=global_display, save_png='PEEF_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_themis_esa_copy_specplot(self):
        del_data("*")
        import pyspedas
        # THEMIS ESA has monotonically decreasing energies, time varying energies, and also has fill
        esa_vars = pyspedas.projects.themis.esa(trange=['2016-07-23', '2016-07-24'], probe='a')
        timespan('2016-07-23',1,'days')
        tplot_options('title', 'Decreasing and time-varying energies, fillvals, should render correctly')
        pyspedas.tplot_copy('tha_peef_en_eflux',new_name='tha_peef_copy')
        tplot('tha_peef_copy', display=global_display, save_png='PEEF_copy_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_erg_specplot(self):
        del_data("*")
        import pyspedas
        # ERG specplots, only vertical lines on the bottom panel for original resample...
        pyspedas.projects.erg.hep(trange=['2017-03-27', '2017-03-28'])
        timespan('2017-03-27',1,'days')
        tplot_options('title', 'Time varying spectral bins, should render correctly')
        tplot(['erg_hep_l2_FEDO_L', 'erg_hep_l2_FEDO_H'], display=global_display, save_png='ERG_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_maven_specplot(self):
        del_data("*")
        from pyspedas.projects.maven.spdf import load
        sta_vars = load(trange=['2020-12-30', '2020-12-31'], instrument='static', datatype='c0-64e2m')
        print(sta_vars)
        timespan('2020-12-30',1,'days')
        # This variable contains all zeroes, and is set to plot with log scaling
        tplot_options('title', 'Should be all the same color')
        tplot('bkg',display=global_display,save_png='MAVEN_test')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    #@unittest.skip(reason="Failing until we establish a default for spec_dim_to_plot")
    def test_maven_fluxes_specplot(self):
        del_data("*")
        from pyspedas.projects.maven.spdf import load
        swe_vars = load(trange=['2014-10-18', '2014-10-19'], instrument='swea')
        print(swe_vars)
        timespan('2014-10-18',1,'days')
        # This variable has 3 dimensions but is not marked in the CDF as being a specplot.
        # This used to crash in reduce_spec_dataset because the spec_dim_to_plot option was missing.
        tplot_options('title', 'Spec data plotted as lines')
        tplot('diff_en_fluxes', display=global_display, save_png='MAVEN_fluxes_test_nospec')
        options('diff_en_fluxes',"spec",1)
        # Setting the "spec" option also sets the spec_dim_to_plot option to v2 in this case
        tplot_options('title', 'Plotting as spectrum with default spec_dim_to_plot (v2)')
        tplot('diff_en_fluxes',display=global_display,save_png='MAVEN_fluxes_test_v2')
        # Test that the "v1" option also works (it used to crash looking for "v" and not checking "v1")
        options('diff_en_fluxes','spec_dim_to_plot',"v1")
        tplot_options('title', 'Plotting as spectrum with spec_dim_to_plot=v1')
        tplot('diff_en_fluxes',display=global_display,save_png='MAVEN_fluxes_test_v1')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_pseudovars_title(self):
        del_data("*")
        import pyspedas
        from pytplot import store_data
        pyspedas.projects.themis.state(probe='c',trange=default_trange)
        store_data('ps1', ['thc_spin_initial_delta_phi', 'thc_spin_idpu_spinper'])
        store_data('ps2', ['thc_spin_initial_delta_phi', 'thc_spin_idpu_spinper'])
        store_data('ps3', ['thc_spin_initial_delta_phi', 'thc_spin_idpu_spinper'])
        tplot_options('title', 'Plot title should only appear once at top of plot')
        plotvars = ['thc_pos', 'ps1', 'thc_vel', 'ps2', 'thc_pos_gse', 'ps3']
        # Should have only one title at the top of the plot
        tplot(plotvars, save_png='test_pseudovars_title', display=global_display)
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_pseudovars_spectra(self):
        del_data("*")
        import pyspedas
        from pytplot import zlim, ylim, timespan

        # Load ESA and SST data
        pyspedas.projects.themis.esa(probe='a', trange=default_trange)
        pyspedas.projects.themis.sst(probe='a', trange=default_trange)

        # Make a combined variable with both ESA and SST spectral data (disjoint energy ranges)
        store_data('combined_spec', ['tha_peif_en_eflux', 'tha_psif_en_eflux'])
        options('tha_peif_en_eflux', 'y_no_resample', 1)
        options('combined_spec','y_range',[5.0, 7e+06])
        vars = ['tha_peif_en_eflux', 'combined_spec', 'tha_psif_en_eflux']
        tplot_options('title', 'Pseudovar with two spectra, disjoint energies: top=ESA, middle=combined, bottom=SST')
        tplot(vars, save_png='test_pseudo_spectra_disjoint_energies', display=global_display)

        tplot_options('title','Original burst variable')
        tplot('tha_peib_en_eflux',display=global_display)
        degap('tha_peib_en_eflux',dt=4.0)
        tplot_options('title','Degapped burst variable')
        tplot('tha_peib_en_eflux',display=global_display)
        # Make a combined variable with full & burst data (same energy ranges, intermittent burst data at higher cadence)
        store_data('esa_srvy_burst', ['tha_peif_en_eflux', 'tha_peib_en_eflux'])
        zlim('tha_peif_en_eflux', 1.0e3, 1.0e7)
        options('tha_peif_en_eflux', 'y_range', [0.5, 1.0e6])

        options('tha_peib_en_eflux', 'y_no_resample', 1)
        zlim('tha_peib_en_eflux', 1.0e3, 1.0e7)
        options('tha_peib_en_eflux', 'y_range', [0.5, 1.0e6])
        options('tha_peib_en_eflux', 'data_gap', 4.0)
        vars = ['tha_peif_en_eflux', 'esa_srvy_burst', 'tha_peib_en_eflux']
        tplot_options('title', 'Combining full & burst cadence with same energies: top=fast, middle=combined, bot=burst')
        tplot(vars, save_png='test_pseudo_spectra_full_burst',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

        # Zoom in o a burst interval
        timespan('2007-03-23/12:20', 10, 'minutes')
        tplot_options('title', 'Combined full and burst cadence with same energies (zoomed in) top=fast, mid=combined, bot=burst')
        tplot(vars, save_png='test_pseudo_spectra_full_burst_zoomed',display=global_display)
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests
        tplot_options('title', '')

    def test_pseudo_spectra_plus_line(self):
        del_data("*")
        import pyspedas
        pyspedas.projects.mms.fpi(datatype='des-moms', trange=['2015-10-16', '2015-10-17'])
        pyspedas.projects.mms.edp(trange=['2015-10-16', '2015-10-17'], datatype='scpot')
        # Create a pseudovariable with an energy spectrum plus a line plot of spacecraft potential
        store_data('spec', data=['mms1_des_energyspectr_omni_fast', 'mms1_edp_scpot_fast_l2', 'mms1_edp_scpot_fast_l2'])
        # Set some options so that the spectrum, trace, and y axes are legible
        options('mms1_edp_scpot_fast_l2', 'yrange', [10, 100])
        #options('mms2_edp_scpot_fast_l2', 'right_axis', True)
        options('spec','right_axis','True')
        tplot_options('xmargin', [0.1, 0.2])
        timespan('2015-10-16',1,'days')
        tplot_options('title', 'Pseudovar with energy spectrum plus line plot of s/c potential')
        tplot('spec', xsize=12, display=global_display,save_png='MMS_pseudo_spec_plus_line')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_pseudo_spectra_plus_line_copy(self):
        del_data("*")
        import pyspedas
        pyspedas.projects.mms.fpi(datatype='des-moms', trange=['2015-10-16', '2015-10-17'])
        pyspedas.projects.mms.edp(trange=['2015-10-16', '2015-10-17'], datatype='scpot')
        # Create a pseudovariable with an energy spectrum plus a line plot of spacecraft potential
        store_data('spec', data=['mms1_des_energyspectr_omni_fast', 'mms1_edp_scpot_fast_l2', 'mms1_edp_scpot_fast_l2'])
        # Set some options so that the spectrum, trace, and y axes are legible
        options('mms1_edp_scpot_fast_l2', 'yrange', [10, 100])
        #options('mms2_edp_scpot_fast_l2', 'right_axis', True)
        options('spec','right_axis','True')
        tplot_options('xmargin', [0.1, 0.2])
        timespan('2015-10-16',1,'days')
        tplot_options('title', 'Pseudovar with energy spectrum plus line plot of s/c potential')
        pyspedas.tplot_copy('spec', 'spec_copy')
        tplot('spec_copy', xsize=12, display=global_display,save_png='MMS_pseudo_spec_plus_line')
        tplot_options('title', '')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_psp_flux_plot(self):
        del_data("*")
        import pyspedas
        import pytplot
        import numpy as np
        # import matplotlib.pyplot as plt
        from pytplot import tplot
        spi_vars = pyspedas.projects.psp.spi(trange=['2022-12-12/00:00', '2022-12-12/23:59'], datatype='sf00_l3_mom', level='l3',
                                    time_clip=True)
        time = pytplot.data_quants['psp_spi_EFLUX_VS_ENERGY'].coords['time'].values
        # print(time)
        energy_channel = pytplot.data_quants['psp_spi_EFLUX_VS_ENERGY'].coords['spec_bins'].values
        # print(energy_channel)
        ec = energy_channel[0, :]
        energy_flux = pytplot.data_quants['psp_spi_EFLUX_VS_ENERGY'].values
        # print(energy_flux)
        e_flux = pytplot.data_quants['psp_spi_EFLUX_VS_ENERGY'].coords['v'].values
        energy_flux[energy_flux == 0] = np.nan
        pytplot.store_data('E_Flux', data={'x': time.T, 'y': energy_flux, 'v': energy_channel})
        pytplot.options('E_Flux', opt_dict={'Spec': 1, 'zlog': 1, 'Colormap': 'jet', 'ylog': 1})
        timespan('2022-12-12',1,'days')
        tplot_options('title', 'Parker Solar Probe E_flux')
        tplot('E_Flux',display=global_display, save_png='psp_E_Flux')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests

    def test_save_with_plot_objects(self):
        # Test that output files are saved when return_plot_objects is True
        import pyspedas
        import os
        from matplotlib import pyplot as plt
        pyspedas.projects.themis.state(probe='a')
        if os.path.exists('ret_plot_objs.png'):
            os.remove('ret_plot_objs.png')
        tplot_options('title', 'Should create ret_plot_objs.png ')
        tplot('tha_pos',display=global_display,return_plot_objects=True, save_png='ret_plot_objs.png')
        self.assertTrue(os.path.exists('ret_plot_objs.png'))

    def test_erg_tplot_vlabels(self):
        # Test alternate varlabel implementation from Tomo Hori
        del_data("*")
        import pyspedas

        from pyspedas.projects.erg import mgf, orb
        timespan('2017-03-27',1,'days')
        mgf()
        orb()
        from pytplot import tplot_options, options, tplot_names, split_vec, get_data, tplot_opt_glob, tnames

        split_vec('erg_orb_l2_pos_rmlatmlt')
        split_vec('erg_orb_l2_pos_Lm')
        options('erg_orb_l2_pos_rmlatmlt_x', 'ytitle', 'R')
        options('erg_orb_l2_pos_rmlatmlt_y', 'ytitle', 'Mlat')
        options('erg_orb_l2_pos_rmlatmlt_z', 'ytitle', 'MLT')

        var_label = ['erg_orb_l2_pos_Lm_x', 'erg_orb_l2_pos_rmlatmlt_x', 'erg_orb_l2_pos_rmlatmlt_y',
                     'erg_orb_l2_pos_rmlatmlt_z']
        # tplot_options('var_label', var_label)

        plot_vars = ['erg_mgf_l2_mag_8sec_sm', 'erg_mgf_l2_igrf_8sec_sm', 'erg_orb_l2_pos_Lm_x']

        from pytplot import tplot_vl
        fig = tplot_vl(plot_vars, var_label=var_label, display=global_display, save_png='erg_varlabel.png')
        timespan('2007-03-23',1,'days') # Reset to avoid interfering with other tests


if __name__ == '__main__':
    unittest.main()



"""Test functions in the utilites folder."""
import unittest

import pyspedas
from pyspedas.utilities.dailynames import dailynames
from pyspedas import tcopy
from pyspedas import themis
from pyspedas import cotrans
from pyspedas import mpause_2, mpause_t96
from pyspedas import find_datasets
from pytplot import data_exists, tkm2re, tplot, split_vec, del_data
from pytplot import get_data, store_data, options
import numpy as np
from numpy.testing import assert_allclose


class UtilTestCases(unittest.TestCase):
    """Test fuctions in the utilites folder."""

    def test_dailynames(self):
        """Test dailynames function."""
        self.assertTrue(dailynames(trange=['2015-12-1', '2015-12-1/2:00'],
                                   hour_res=True) == ['2015120100',
                                                      '2015120101'])
        self.assertTrue(dailynames(trange=['2015-12-1', '2015-12-3'])
                        == ['20151201', '20151202'])
        self.assertTrue(dailynames(trange=['2015-12-3', '2015-12-2'])
                        == ['20151203'])
        self.assertTrue(dailynames() is None)
        self.assertTrue(dailynames(trange=['2015-12-3', '2019-12-2'],
                                   file_format='%Y') ==
                        ['2015', '2016', '2017', '2018', '2019'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-3-2'],
                                   file_format='%Y%m') == ['201501', '201502',
                                                           '201503'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-3-2'],
                                   file_format='/%Y/%m/') ==
                        ['/2015/01/', '/2015/02/', '/2015/03/'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-1-1/3:00'],
                                   file_format='%H', res=60.0) ==
                        ['00', '01', '02'])
        self.assertTrue(dailynames(trange=['2015-1-1/2:00', '2015-1-1/3:00'],
                                   file_format='%M', res=600.) ==
                        ['00', '10', '20', '30', '40', '50'])

    def test_tcopy(self):
        """Test tcopy function."""
        store_data('test', data={'x': [1, 2, 3], 'y': [5, 5, 5]})
        tcopy('test')
        tcopy('test', 'another-copy')
        t, d = get_data('test-copy')
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        t, d = get_data('another-copy')
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        # the following should gracefully error
        tcopy('doesnt exist', 'another-copy')
        tcopy(['another-copy', 'test'], 'another-copy')

    def test_tkm2re(self):
        store_data('test', data={'x': [1, 2, 3], 'y': [5, 5, 5]})
        options('test', 'ysubtitle', '[Re]')
        # convert to km
        tkm2re('test', km=True)
        # convert back
        tkm2re('test_km')
        self.assertTrue(data_exists('test_km_re'))
        nothing = tkm2re('doesnt_exist')
        self.assertTrue(nothing is None)
        tkm2re('test_km', newname='another_test_km')
        self.assertTrue(data_exists('another_test_km'))
        anerror = tkm2re('test_km', newname=['test1_km', 'test1_km'])
        self.assertTrue(anerror is None)

    def test_time_clip(self):
        import pytplot
        x=[1,2,3,4,5]
        y=[2,4,6,8,10]
        xfp = [1.0,2.0,3.0,4.0,5.0]
        # Data for a scalar variable
        y_scalar = [0.0, 1.0, 2.0, 3.0, 4.0]
        # Data for a vector variable
        y_vec = [[0.0, 1.0, 2.0],[3.0,4.0,5.0],[6.0,7.0,8.0],[9.0,10.0,11.0],[12.0,13.0,14.0]]
        y_vec = np.array(y_vec)
        # Constant metadata for a vector variable
        v1_1d = [1.0, 2.0, 3.0]
        # Time varying metadata for a vector variable
        v1_2d = y_vec
        # Data for a 3x3 matrix variable
        y_mat = [[[0.0, 1.0, 2.0],    [0.0, 1.0, 2.0],  [0.0, 1.0, 2.0]],
                 [[3.0, 4.0, 5.0],    [3.0,4.0,5.0],    [3.0,4.0,5.0]],
                 [[6.0, 7.0, 8.0],    [6.0,7.0,8.0],    [6.0,7.0,8.0]],
                 [[9.0, 10.0, 11.0],  [9.0,10.0,11.0],  [9.0,10.0,11.0]],
                 [[12.0, 13.0, 14.0], [12.0,13.0,14.0], [12.0,13.0,14.0]]]
        y_mat = np.array(y_mat)
        v2_1d = v1_1d
        v2_2d = y_vec

        # Scalar
        pytplot.store_data('fptest',data={'x':xfp,'y':y_scalar})
        # 1-D vec, no metadata
        pytplot.store_data('fptest_vec_no_v1',data={'x':xfp,'y':y_vec})
        # 1-D vec, with 1-D (constant) V1
        pytplot.store_data('fptest_vec_v1_1d',data={'x':xfp,'y':y_vec, 'v':v1_1d})
        # 1-D vec, with 2-D (time varying) V1
        pytplot.store_data('fptest_vec_v1_2d',data={'x':xfp,'y':y_vec, 'v':v1_2d})
        # 3x3 mat, with no V1, no V2
        pytplot.store_data('fptest_mat_v1_none_v2_none',data={'x':xfp,'y':y_mat})
        # 3x3 mat, with 1-D (constant) V1, 1-D (constant) V2
        pytplot.store_data('fptest_mat_v1_1d_v2_1d',data={'x':xfp,'y':y_mat, 'v1':v1_1d, 'v2':v2_1d})
        # 3x3 mat, with 2-D (time varying) V1, 2-D (time varying) V2
        pytplot.store_data('fptest_mat_v1_2d_v2_2d',data={'x':xfp,'y':y_mat, 'v1':v1_2d, 'v2':v2_2d})
        # 3x3 mat, with 2-D (time varying) V1, 1-D (constant) V2
        pytplot.store_data('fptest_mat_v1_2d_v2_1d',data={'x':xfp,'y':y_mat, 'v1':v1_2d, 'v2':v2_1d})
        # 3x3 mat, with 1-D (constant) V1, 2-D (time varying) V2
        pytplot.store_data('fptest_mat_v1_1d_v2_2d',data={'x':xfp,'y':y_mat, 'v1':v1_1d, 'v2':v2_2d})
        # 3x3 mat, with missing V1, 2-D (time varying) V2
        # This doesn't quite work.  The 'v2' value doesn't survive the store_data/get_data round trip
        # if v1 is missing.  Not sure if that really qualifies as a bug, considering how ill-advised such a
        # construct would be...
        # pytplot.store_data('fptest_mat_v1_none_v2_2d',data={'x':xfp,'y':y_mat, 'v1':None, 'v2':v2_2d})



        # Test warning for no data in time range
        pytplot.time_clip('fptest',1.5,1.7)

        # Single value in time range
        pytplot.time_clip('fptest',1.5,2.5)
        self.assertTrue(data_exists('fptest-tclip'))
        dat=get_data('fptest-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)

        # Multiple values in time range
        pytplot.time_clip('fptest',1.5,3.5)
        dat=get_data('fptest-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])

        # Multiple values in time range -- clip interior
        pytplot.time_clip('fptest',1.5,3.5, suffix='-tclip-interior', interior_clip=True)
        dat=get_data('fptest-tclip-interior')
        self.assertTrue(len(dat.times) == 3)
        assert_allclose(dat.times, [1.0, 4.0, 5.0])

        # Entire time range
        pytplot.time_clip('fptest',-100.0,100.0, suffix='-full_range')
        self.assertTrue(data_exists('fptest-full_range'))
        dat=get_data('fptest-full_range')
        self.assertTrue(len(dat.times) == 5)
        assert_allclose(dat.times, xfp)

        # Vector data, single value in time range, no metadata
        pytplot.time_clip('fptest_vec_no_v1',1.5,2.5)
        self.assertTrue(data_exists('fptest_vec_no_v1-tclip'))
        dat=get_data('fptest_vec_no_v1-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0,:], [3.0,4.0,5.0])

        # Vector data, single value in time range, constant metadata
        pytplot.time_clip('fptest_vec_v1_1d',1.5,2.5)
        self.assertTrue(data_exists('fptest_vec_v1_1d-tclip'))
        dat=get_data('fptest_vec_v1_1d-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0,:], [3.0,4.0,5.0])
        assert_allclose(dat.v, v1_1d)

        # Vector data, multiple values in time range, constant metadata
        pytplot.time_clip('fptest_vec_v1_1d',1.5,3.5)
        self.assertTrue(data_exists('fptest_vec_v1_1d-tclip'))
        dat=get_data('fptest_vec_v1_1d-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        assert_allclose(dat.y, [[3.0,4.0,5.0], [6.0, 7.0, 8.0]])
        assert_allclose(dat.v, v1_1d)

        # Vector data, single value in time range, time-varying metadata
        pytplot.time_clip('fptest_vec_v1_2d',1.5,2.5)
        self.assertTrue(data_exists('fptest_vec_v1_2d-tclip'))
        dat=get_data('fptest_vec_v1_2d-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v.shape == (1,3))
        assert_allclose(dat.y[0,:], [3.0,4.0,5.0])
        assert_allclose(dat.v, v1_2d[1:2])

        # Vector data, multiple values in time range, time-varying metadata
        pytplot.time_clip('fptest_vec_v1_2d',1.5,3.5)
        self.assertTrue(data_exists('fptest_vec_v1_2d-tclip'))
        dat=get_data('fptest_vec_v1_2d-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, xfp[1:3])
        self.assertTrue(dat.y.shape == (2,3))
        assert_allclose(dat.y, y_vec[1:3,:])
        self.assertTrue(dat.v.shape == (2,3))
        assert_allclose(dat.v, v1_2d[1:3])

        # matrix data, single value in time range, no metadata
        pytplot.time_clip('fptest_mat_v1_none_v2_none',1.5,2.5)
        self.assertTrue(data_exists('fptest_mat_v1_none_v2_none-tclip'))
        dat=get_data('fptest_mat_v1_none_v2_none-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0,:,:], y_mat[1,:,:])

        # matrix data, single value in time range, constant v1 and v2 metadata
        pytplot.time_clip('fptest_mat_v1_1d_v2_1d',1.5,2.5)
        self.assertTrue(data_exists('fptest_mat_v1_1d_v2_1d-tclip'))
        dat=get_data('fptest_mat_v1_1d_v2_1d-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0,:,:], y_mat[1,:,:])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v2_1d)

        # matrix data, multiple values in time range, constant v1 and v2 metadata
        pytplot.time_clip('fptest_mat_v1_1d_v2_1d',1.5,3.5)
        self.assertTrue(data_exists('fptest_mat_v1_1d_v2_1d-tclip'))
        dat=get_data('fptest_mat_v1_1d_v2_1d-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        assert_allclose(dat.y, y_mat[1:3,:,:])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v2_1d)

        # matrix data, single value in time range, constant v1, time-varying v2 metadata
        pytplot.time_clip('fptest_mat_v1_1d_v2_2d',1.5,2.5)
        self.assertTrue(data_exists('fptest_mat_v1_1d_v2_2d-tclip'))
        dat=get_data('fptest_mat_v1_1d_v2_2d-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v1.shape == (3,))
        assert_allclose(dat.y[:,:,:], y_mat[1:2,:,:])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v1_2d[1:2])

        # matrix data, multiple values in time range, constant v1, time-varying v2 metadata
        pytplot.time_clip('fptest_mat_v1_1d_v2_2d',1.5,3.5)
        self.assertTrue(data_exists('fptest_mat_v1_1d_v2_2d-tclip'))
        dat=get_data('fptest_mat_v1_1d_v2_2d-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, xfp[1:3])
        self.assertTrue(dat.v1.shape == (3,))
        assert_allclose(dat.y[:,:,:], y_mat[1:3,:,:])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v1_2d[1:3])

        # matrix data, single values in time range, time-varying v1, constant v2 metadata
        pytplot.time_clip('fptest_mat_v1_2d_v2_1d',1.5,2.5)
        self.assertTrue(data_exists('fptest_mat_v1_2d_v2_1d-tclip'))
        dat=get_data('fptest_mat_v1_2d_v2_1d-tclip')
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v1.shape == (1,3))
        self.assertTrue(dat.v2.shape == (3,))
        assert_allclose(dat.y[:,:,], y_mat[1:2,:,:])
        assert_allclose(dat.v1, v1_2d[1:2])
        assert_allclose(dat.v2, v2_1d)

        # matrix data, multiple values in time range, time-varying v1, constant v2 metadata
        pytplot.time_clip('fptest_mat_v1_2d_v2_1d',1.5,3.5)
        self.assertTrue(data_exists('fptest_mat_v1_2d_v2_1d-tclip'))
        dat=get_data('fptest_mat_v1_2d_v2_1d-tclip')
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        self.assertTrue(dat.v1.shape == (2,3))
        self.assertTrue(dat.v2.shape == (3,))
        assert_allclose(dat.y[:,:,], y_mat[1:3,:,:])
        assert_allclose(dat.v1, v1_2d[1:3])
        assert_allclose(dat.v2, v2_1d)

        # Multiple variables in one call
        pytplot.store_data('tst1',data={'x':x,'y':y})
        pytplot.store_data('tst2',data={'x':x,'y':y})
        # reversed time limits
        pytplot.time_clip(['tst1','tst2'],10,1)
        # data completely outside time limits
        pytplot.time_clip(['tst1','tst2'],10,20)
        # one point in limits
        pytplot.time_clip('tst1',1.5,2.5)
        self.assertTrue(data_exists('tst1-tclip'))
        # overwrite
        pytplot.del_data('tst1-tclip')
        pytplot.time_clip('tst1',1.5,2.0,overwrite=True)
        self.assertFalse(data_exists('tst1-tclip'))
        pytplot.time_clip('tst1',1.5,2.5,newname='tst1_new')
        self.assertTrue(data_exists('tst1_new'))
        pytplot.time_clip('tst1',1.5,2.5,newname='',suffix='-tc1')
        self.assertTrue(data_exists('tst1-tc1'))
        pytplot.time_clip('tst1',1.5,2.5,newname=[],suffix='-tc2')
        self.assertTrue(data_exists('tst1-tc2'))
        # newname has different count than input names, default to suffix
        pytplot.time_clip('tst1',1.5,2.5, newname=['foo','bar'],suffix='-tc3')
        self.assertTrue(data_exists('tst1-tc3'))
        pytplot.time_clip('tst1',1.5,2.5,newname=None, suffix='-tc4')
        self.assertTrue(data_exists('tst1-tc4'))
        # no such tplot name
        pytplot.time_clip('bogus',1.5,2.5)
        # empty input list
        pytplot.time_clip([],1.5,2.5)

        # Strings as start/end times
        trange = ['2007-03-23/9:30', '2007-03-23/12:00']
        # Note times are out of order if compared as strings!
        self.assertTrue(trange[0] > trange[1])
        pyspedas.projects.themis.state(probe='a',trange=trange)
        pyspedas.time_clip('tha_pos',trange[0], trange[1],newname='tha_pos-tclip')
        self.assertTrue(data_exists('tha_pos-tclip'))
        dat=get_data('tha_pos-tclip')
        self.assertTrue(dat.times[0] >= pyspedas.time_float(trange[0]))


    def test_mpause_t96(self):
        trange = ['2023-03-24', '2023-03-25']
        # THEMIS orbits come from the 'state' datatype
        # We will use GSE coordinates for the plots we'll make.
        themis.state(probe='a', trange=trange, varformat='*_pos_gse')
        cotrans('tha_pos_gse', name_out='tha_pos_gsm', coord_in='GSE', coord_out='GSM')
        posdat = get_data('tha_pos_gsm')
        re = 6378.0
        pos_gsm = posdat.y / re
        xmgnp, ymgnp, zmgnp, id, distan = mpause_t96(
            pd=14.0, xgsm=pos_gsm[:, 0], ygsm=pos_gsm[:, 1], zgsm=pos_gsm[:, 2])
        self.assertEqual(len(xmgnp), 90)
        self.assertEqual(id.min(), -1)
        self.assertEqual(id.max(), 1)
        self.assertTrue(distan.min() < 0.01)
        self.assertTrue(distan.max() > 10.0)
        self.assertEqual(len(id), len(posdat.times))

    def test_mpause_2(self):
        trange = ['2023-03-24', '2023-03-25']
        # THEMIS orbits come from the 'state' datatype
        # We will use GSE coordinates for the plots we'll make.
        themis.state(probe='a', trange=trange, varformat='*_pos_gse')
        cotrans('tha_pos_gse', name_out='tha_pos_gsm', coord_in='GSE', coord_out='GSM')
        posdat = get_data('tha_pos_gsm')
        re = 6378.0
        pos_gsm = posdat.y / re
        xmp, ymp  = mpause_2()
        self.assertEqual(len(xmp),2000)
        self.assertEqual(len(ymp), 2000)
        self.assertTrue(xmp.min() < -299.0)
        self.assertTrue(xmp.max() > 10.77)
        self.assertTrue(ymp.min() < -68.0)
        self.assertTrue(ymp.max() > 77.8)

    def test_find_datasets_quiet(self):
        ds = find_datasets(mission='MMS', instrument='FGM', quiet=True)
        self.assertTrue('MMS1_FGM_BRST_L2' in ds)
        self.assertTrue('MMS2_FGM_BRST_L2' in ds)

    def test_find_datasets_nolabel(self):
        ds = find_datasets(mission='MMS', instrument='FGM')
        self.assertTrue('MMS1_FGM_BRST_L2' in ds)
        self.assertTrue('MMS2_FGM_BRST_L2' in ds)

    def test_find_datasets_label(self):
        ds = find_datasets(mission='MMS', instrument='FGM',label=True)
        self.assertTrue('MMS1_FGM_BRST_L2' in ds)
        self.assertTrue('MMS2_FGM_BRST_L2' in ds)

    def test_split_vec_metadata(self):
        del_data('*')
        pyspedas.projects.themis.fit(probe='c')
        md = get_data('thc_fgs_dsl')
        split_vec('thc_fgs_dsl')
        self.assertTrue(data_exists('thc_fgs_dsl_x'))
        self.assertTrue(data_exists('thc_fgs_dsl_y'))
        self.assertTrue(data_exists('thc_fgs_dsl_z'))
        md_x = get_data('thc_fgs_dsl_x', metadata=True)
        self.assertTrue(md_x['plot_options']['xaxis_opt']['axis_label'] == 'Time')

    def test_imports(self):
        import pyspedas
        from pytplot import del_data, data_exists, tplot_names
        # fully qualified name without .projects.
        del_data('*')
        pyspedas.projects.themis.state(probe='a')
        self.assertTrue(data_exists('tha_pos'))
        # fully qualified name with .projects.
        del_data('*')
        pyspedas.projects.themis.state(probe='a')
        self.assertTrue(data_exists('tha_pos'))

        # import themis without .projects.
        from pyspedas import themis
        del_data('*')
        themis.state(probe='a')
        self.assertTrue(data_exists('tha_pos'))
        # import themis with .projects.
        from pyspedas.projects import themis
        del_data('*')
        themis.state(probe='a')
        self.assertTrue(data_exists('tha_pos'))

        # import themis.state without .projects.
        # PyCharm's static analysis doesn't like this (red underlines) but it works at runtime
        from pyspedas.themis import state
        del_data('*')
        state(probe='a')
        self.assertTrue(data_exists('tha_pos'))
        # import state with .projects.
        from pyspedas.projects.themis import state
        del_data('*')
        state(probe='a')
        self.assertTrue(data_exists('tha_pos'))

        from pyspedas import mms
        del_data('*')
        mms.fgm()
        self.assertTrue(data_exists('mms1_fgm_b_gsm_srvy_l2'))

        from pyspedas.projects.mms import fgm
        del_data('*')
        fgm()
        self.assertTrue(data_exists('mms1_fgm_b_gsm_srvy_l2'))

        # deep import themis.state without .projects.
        from pyspedas.themis.state_tools.state import state
        del_data('*')
        state(probe='a')
        self.assertTrue(data_exists('tha_pos'))
        # deep import themis.state with .projects.
        from pyspedas.projects.themis.state_tools.state import state
        del_data('*')
        state(probe='a')
        self.assertTrue(data_exists('tha_pos'))

if __name__ == '__main__':
    unittest.main()



import os
import requests
import logging
import warnings
from importlib.metadata import version, PackageNotFoundError
import numpy as np
from pytplot import cdf_to_tplot
from pytplot import time_clip as tclip
from pytplot import time_double, time_string
from dateutil.parser import parse
from datetime import timedelta, datetime
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from .mms_config import CONFIG
from .mms_get_local_files import mms_get_local_files
from .mms_files_in_interval import mms_files_in_interval
from .mms_login_lasp import mms_login_lasp
from .mms_file_filter import mms_file_filter
from .mms_load_data_spdf import mms_load_data_spdf

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def mms_load_data(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', level='l2', 
    instrument='fgm', datatype='', varformat=None, exclude_format=None, prefix='', suffix='', get_support_data=False, time_clip=False,
    no_update=False, center_measurement=False, available=False, notplot=False, latest_version=False, 
    major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False, varnames=[]):
    """
    This function loads MMS data into pyTplot variables

    This function is not meant to be called directly. Please see the individual load routines for documentation and use.
    """
    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]
    if not isinstance(datatype, list): datatype = [datatype]
    
    probe = [str(p) for p in probe]

    # allows the user to pass in trange as list of datetime objects
    if type(trange[0]) == datetime and type(trange[1]) == datetime:
        trange = [time_string(trange[0].timestamp()), time_string(trange[1].timestamp())]

    # allows the user to pass in trange as a list of floats (unix times)
    if isinstance(trange[0], float):
        trange[0] = time_string(trange[0])
    if isinstance(trange[1], float):
        trange[1] = time_string(trange[1])

    download_only = CONFIG['download_only']

    no_download = False
    if no_update or CONFIG['no_download']: no_download = True

    if spdf:
        return mms_load_data_spdf(trange=trange, probe=probe, data_rate=data_rate, level=level, 
                                  instrument=instrument, datatype=datatype, varformat=varformat, exclude_format=exclude_format,
                                  suffix=suffix, get_support_data=get_support_data, time_clip=time_clip, 
                                  no_update=no_update, center_measurement=center_measurement, notplot=notplot, 
                                  latest_version=latest_version, major_version=major_version, 
                                  min_version=min_version, cdf_version=cdf_version, varnames=varnames)

    headers = {}
    try:
        release_version = version("pyspedas")
    except PackageNotFoundError:
        release_version = "bleeding edge"
    headers['User-Agent'] = 'pySPEDAS ' + release_version

    user = None
    if not no_download:
        sdc_session, user = mms_login_lasp(always_prompt=always_prompt, headers=headers)

    out_files = []
    available_files = []

    # We want to load CDFs by groups (probe, level, datatype, data rate) rather than
    # all together in one giant call to cdf_to_tplot.  We'll use a dictionary, with keys
    # derived from each combination of attributes, and maintain separate lists of files to
    # load together for each key.

    out_file_groupings = {}

    for prb in probe:
        for drate in data_rate:
            start_date = parse(trange[0]).strftime('%Y-%m-%d') # need to request full day, then parse out later
            end_date = parse(time_string(time_double(trange[1])-0.1)).strftime('%Y-%m-%d-%H-%M-%S') # -1 second to avoid getting data for the next day
            # kludge to fix issue for burst mode data in files from the previous day
            if drate == 'brst':
                sec_from_start_of_day = time_double(trange[0])-time_double(start_date)

                # check if we're within 10 minutes of the start of the day
                # and if so, grab 10 minutes of data from the end of the
                # previous day
                if sec_from_start_of_day <= 600.0:
                    start_date = time_string(time_double(start_date)-600.0, fmt='%Y-%m-%d-%H-%M-%S')

            for lvl in level:
                for dtype in datatype:

                    grouping_key = f"probe: {prb}, drate: {drate}, level: {lvl}, datatype: {dtype}"
                    out_file_groupings[grouping_key] = []

                    file_found = False

                    if user is None:
                        url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/file_info/science?start_date=' + start_date + '&end_date=' + end_date + '&sc_id=mms' + prb + '&instrument_id=' + instrument + '&data_rate_mode=' + drate + '&data_level=' + lvl
                    else:
                        url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/file_info/science?start_date=' + start_date + '&end_date=' + end_date + '&sc_id=mms' + prb + '&instrument_id=' + instrument + '&data_rate_mode=' + drate + '&data_level=' + lvl
                    
                    if dtype != '':
                        url = url + '&descriptor=' + dtype

                    if CONFIG['debug_mode']: logging.info('Fetching: ' + url)

                    if not no_download:
                        # query list of available files
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore", category=ResourceWarning)
                                http_request = sdc_session.get(url, verify=True, headers=headers)
                                if http_request.status_code != 200:
                                    logging.warning("Request to MMS SDC returned HTTP status code %d", http_request.status_code)
                                    logging.warning("Text: %s", http_request.text)
                                    logging.warning("URL: %s", url)
                                    continue
                                else:
                                    http_json = http_request.json()

                            if CONFIG['debug_mode']: logging.info('Filtering the results down to your trange')

                            files_in_interval = mms_files_in_interval(http_json['files'], trange)

                            if available:
                                for file in files_in_interval:
                                    logging.info(file['file_name'] + ' (' + str(np.round(file['file_size']/(1024.*1024), decimals=1)) + ' MB)')
                                    available_files.append(file['file_name'])
                                continue

                            for file in files_in_interval:
                                file_date = parse(file['timetag'])
                                sep = "/" if is_fsspec_uri(CONFIG["local_data_dir"]) else os.path.sep
                                if dtype == '':
                                    out_dir = sep.join([CONFIG['local_data_dir'], 'mms'+prb, instrument, drate, lvl, file_date.strftime('%Y'), file_date.strftime('%m')])
                                else:
                                    out_dir = sep.join([CONFIG['local_data_dir'], 'mms'+prb, instrument, drate, lvl, dtype, file_date.strftime('%Y'), file_date.strftime('%m')])

                                if drate.lower() == 'brst':
                                    out_dir = sep.join([out_dir, file_date.strftime('%d')])

                                out_file = sep.join([out_dir, file['file_name']])

                                if CONFIG['debug_mode']: logging.info('File: ' + file['file_name'] + ' / ' + file['timetag'])

                                if is_fsspec_uri(CONFIG["local_data_dir"]):
                                    protocol, path = out_file.split("://")
                                    fs = fsspec.filesystem(protocol)

                                    if fs.exists(out_file) and str(fs.size(out_file)) == str(file["file_size"]):
                                        if not download_only: logging.info('Streaming ' + out_file)
                                        out_files.append(out_file)
                                        out_file_groupings[grouping_key].append(out_file)
                                        file_found = True
                                        continue
                                else:
                                    if os.path.exists(out_file) and str(os.stat(out_file).st_size) == str(file['file_size']):
                                        if not download_only: logging.debug('Loading ' + out_file)
                                        out_files.append(out_file)
                                        out_file_groupings[grouping_key].append(out_file)
                                        file_found = True
                                        continue

                                if user is None:
                                    download_url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/download/science?file=' + file['file_name']
                                else:
                                    download_url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/download/science?file=' + file['file_name']

                                logging.info('Downloading ' + file['file_name'] + ' to ' + out_dir)

                                with warnings.catch_warnings():
                                    warnings.simplefilter("ignore", category=ResourceWarning)
                                    fsrc = sdc_session.get(download_url, stream=True, verify=True, headers=headers)
                                ftmp = NamedTemporaryFile(delete=False)

                                with open(ftmp.name, 'wb') as f:
                                    copyfileobj(fsrc.raw, f)

                                if is_fsspec_uri(CONFIG["local_data_dir"]):
                                    protocol, path = out_dir.split("://")
                                    fs = fsspec.filesystem(protocol)

                                    fs.makedirs(out_dir, exist_ok=True)

                                    # if the download was successful, put at URI specified
                                    fs.put(ftmp.name, out_file)
                                else:
                                    if not os.path.exists(out_dir):
                                        os.makedirs(out_dir)

                                    # if the download was successful, copy to data directory
                                    copy(ftmp.name, out_file)

                                out_files.append(out_file)
                                out_file_groupings[grouping_key].extend([out_file])
                                file_found = True
                                fsrc.close()
                                ftmp.close()
                                os.unlink(ftmp.name)  # delete the temporary file
                        except requests.exceptions.ConnectionError as e:
                            # No/bad internet connection; try loading the files locally
                            print(e)
                            logging.error('No internet connection!')

                    if not file_found:
                        added_local_files = False
                        if not download_only:
                            logging.info('Searching for local files...')
                            local_files = mms_get_local_files(prb, instrument, drate, lvl, dtype, trange)
                            out_files.extend(local_files)
                            out_file_groupings[grouping_key].extend(local_files)
                            added_local_files = True

                        if added_local_files and CONFIG['mirror_data_dir'] is not None:
                            # check for network mirror; note: network mirrors are assumed to be read-only
                            # and we always copy the files from the mirror to the local data directory
                            # before trying to load into tplot variables 
                            logging.info('No local files found; checking network mirror...')
                            local_files = mms_get_local_files(prb, instrument, drate, lvl, dtype, trange, mirror=True)
                            out_files.extend(local_files)
                            out_file_groupings[grouping_key].extend(local_files)

    if not no_download:
        sdc_session.close()

    if available:
        return available_files

    if not download_only:
        out_files = sorted(out_files)

        # We could be returning a dict or a list, depending on whether notplot is set.
        # In either case, we probably want to eliminate duplicates.  So depending on
        # notplot, we create an empty dict or set, update() it with each returned dictionary (notplot)
        # or list (default), then if we're doing tplot variables, convert it back to a list
        # after processing all the groups.

        if notplot:
            return_value = {}
        else:
            return_value = set()

        for key in out_file_groupings.keys():
            group_list = out_file_groupings[key]
            if len(group_list) > 0:
                sorted_group_list = sorted(group_list)
                filtered_group_list = mms_file_filter(sorted_group_list, latest_version=latest_version, major_version=major_version, min_version=min_version, version=cdf_version)
                if not filtered_group_list:
                    logging.info(f"No matching CDF versions found for group: {key}, after sorting and filtering.")
                else:
                    logging.info(f"Loading files for group: {key}, after sorting and filtering:")
                    for f in filtered_group_list:
                        logging.info(f)
                    these_variables = cdf_to_tplot(filtered_group_list, varformat=varformat,exclude_format=exclude_format, varnames=varnames, get_support_data=get_support_data, prefix=prefix, suffix=suffix, center_measurement=center_measurement, notplot=notplot)
                    return_value.update(these_variables)

        filtered_out_files = mms_file_filter(out_files, latest_version=latest_version, major_version=major_version, min_version=min_version, version=cdf_version)

        if not filtered_out_files:
            logging.info('No matching CDF versions found.')
            return

        # new_variables = cdf_to_tplot(filtered_out_files, varformat=varformat,exclude_format=exclude_format, varnames=varnames, get_support_data=get_support_data, prefix=prefix, suffix=suffix, center_measurement=center_measurement, notplot=notplot)

        if notplot:
            return return_value

        new_variables = list(return_value)
        if not new_variables:
            logging.warning('No data loaded.')
            return

        if time_clip:
            for new_var in new_variables:
                tclip(new_var, trange[0], trange[1], suffix='')

        return new_variables
    else:
        return out_files



from .datasets import datasets
from .efw import efw
from .emfisis import emfisis
from .hope import hope
from .mageis import mageis
from .magephem import magephem
from .rbspice import rbspice
from .rept import rept
from .rps import rps





import os

CONFIG = {
    "local_data_dir": "geom_indices/",
    "remote_data_dir": "http://themis.ssl.berkeley.edu/data/themis/",
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join(
        [os.environ["SPEDAS_DATA_DIR"], "geom_indices"]
    )

if os.environ.get("KP_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["KP_DATA_DIR"]



import unittest
from pyspedas import CDAWeb
from pytplot import data_exists, del_data, get_data, time_double, tplot_names, tplot


class CDAWebTests(unittest.TestCase):
    def test_get_observatories(self):
        cdaweb_obj = CDAWeb()
        obslist = cdaweb_obj.get_observatories()
        self.assertTrue('ARTEMIS' in obslist)

    def test_get_instruments(self):
        cdaweb_obj = CDAWeb()
        inst_list = cdaweb_obj.get_instruments()
        self.assertTrue('Magnetic Fields (space)' in inst_list)

    def test_get_datasets(self):
        cdaweb_obj = CDAWeb()
        ds_list = cdaweb_obj.get_datasets(['Voyager'], ['Magnetic Fields (space)'])
        self.assertTrue('VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)' in ds_list)

    def test_get_filenames(self):
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(['VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)'],
                                           "1979-01-01", "1979-01-02")
        self.assertTrue('https://cdaweb.gsfc.nasa.gov/sp_phys/data/voyager/voyager2/magnetic_fields_cdaweb/mag_2s/1979/voyager2_2s_mag_19790101_v01.cdf' in urllist)

    def test_load_merge(self):
        del_data('*')
        # Create the CDAWeb interface object
        cdaweb_obj = CDAWeb()

        # We'll pick one of available data sets and load it into tplot variables
        dataset = 'VOYAGER2_COHO1HR_MERGED_MAG_PLASMA'
        start_time = '2009-01-01 00:00:00'
        end_time = '2010-01-01 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)

        # Download the data and load as tplot variables.  Setting a prefix
        # is useful if you want to work with both Voyager 1 and Voyager 2
        # data; the variable names in the archived data are the same for both
        # spacecraft.

        cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='v2_')
        dat = get_data('v2_protonDensity')
        # Files are monthly; we should have over a year of data.
        self.assertTrue((dat.times[-1] - dat.times[0])/86400.0 > 395.0)

    def test_load_icon_netcdf(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist, "cdaweb/")
        self.assertTrue(data_exists('ICON_L22_Fringe_Amplitude'))

    def test_load_icon_netcdf_default_dir(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['ICON_L2-2_MIGHTI_VECTOR-WIND-GREEN (2019-12-06 to 2022-11-25)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist)
        self.assertTrue(data_exists('ICON_L22_Fringe_Amplitude'))

    def test_load_goes_netcdf(self):
        del_data('*')
        # Create the CDAWeb interface object
        import pyspedas

        cdaweb_obj = pyspedas.CDAWeb()

        time0 = '2021-01-15T14:05:52'
        time1 = '2021-01-15T15:08:57'

        urllist = cdaweb_obj.get_filenames(['DN_MAGN-L2-HIRES_G17 (2018-08-01 to 2023-03-14)'], time0,
                                           time1)
        result = cdaweb_obj.cda_download(urllist, "cdaweb/")
        self.assertTrue(data_exists('b_gse'))


    def test_load_time_clip(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        dataset = 'NEW_HORIZONS_SWAP_VALIDSUM (2008-10-10 to 2023-07-31)'
        start_time = '2014-10-10 00:00:00'
        end_time = '2014-11-10 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='nh_', trange=[start_time, end_time], time_clip=True)

        # Verify that the times were clipped
        # Note that tplot variables with no data in the requested interval will NOT be clipped (deleted).
        dat = get_data('nh_n')
        self.assertTrue(dat.times[0] >= time_double(start_time))
        self.assertTrue(dat.times[-1] <= time_double(end_time))

    def test_load_time_clip_no_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        dataset = 'NEW_HORIZONS_SWAP_VALIDSUM (2008-10-10 to 2023-07-31)'
        start_time = '2014-10-10 00:00:00'
        end_time = '2014-11-10 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='nh_', time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "No trange specified" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_empty_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        dataset = 'NEW_HORIZONS_SWAP_VALIDSUM (2008-10-10 to 2023-07-31)'
        start_time = '2014-10-10 00:00:00'
        end_time = '2014-11-10 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='nh_', trange=[start_time, start_time],  time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "equal" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_time_clip_backward_trange(self):
        del_data('*')
        cdaweb_obj = CDAWeb()

        dataset = 'NEW_HORIZONS_SWAP_VALIDSUM (2008-10-10 to 2023-07-31)'
        start_time = '2014-10-10 00:00:00'
        end_time = '2014-11-10 00:00:00'

        # Get the URLs for the available data in this time range
        urllist = cdaweb_obj.get_filenames([dataset], start_time, end_time)
        with self.assertLogs(level='WARNING') as logs:
            cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='nh_', trange=[end_time, start_time],  time_clip=True)
        got_trange_warning = False
        for o in logs.output:
            if "out of order" in o:
                got_trange_warning = True
        self.assertTrue(got_trange_warning)

    def test_load_data_prefix(self):
        del_data('*')
        cdaweb_obj = CDAWeb()
        urllist = cdaweb_obj.get_filenames(['VOYAGER2_2S_MAG (1977-08-24 to 1991-01-01)'],
                                           "1979-01-01", "1979-01-02")
        cdaweb_obj.cda_download(urllist, "cdaweb/", prefix='v2_')
        self.assertTrue(data_exists('v2_B1'))


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime, timezone
import numpy as np
from pytplot import time_string, time_datetime
from pytplot import time_float, time_double, degap, store_data, get_data, options, data_exists


class TimeTestCases(unittest.TestCase):
    def test_time_datetime(self):
        """Test time_datetime function."""
        now = time_datetime()
        self.assertTrue(time_datetime('2015-12-15/00:00') == datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc))
        self.assertTrue(time_datetime(1450137600.0000000) == datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc))
        self.assertTrue([time_datetime(1450137600.0000000), time_datetime(1444953600.0000000)] 
            == [datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc), datetime(2015, 10, 16, 0, 0, tzinfo=timezone.utc)])
        self.assertTrue(time_datetime([1450137600.0000000, 1444953600.0000000])
            == [datetime(2015, 12, 15, 0, 0, tzinfo=timezone.utc), datetime(2015, 10, 16, 0, 0, tzinfo=timezone.utc)])


    def test_time_string(self):
        """Test time_string function."""
        self.assertTrue(time_string(fmt='%Y-%m-%d') == datetime.now().strftime('%Y-%m-%d'))
        self.assertTrue(time_string(1450181243.767, fmt='%Y-%m-%d') == '2015-12-15')
        self.assertTrue(time_string(1450181243.767, fmt='%Y-%m-%d/%H:%M:%S') == '2015-12-15/12:07:23')
        self.assertTrue(time_string(1450181243.767) == '2015-12-15 12:07:23.767000')
        self.assertTrue(time_string([1450181243.767, 1450181263.767]) == ['2015-12-15 12:07:23.767000', '2015-12-15 12:07:43.767000'])

    def test_time_double(self):
        """Test time_double function."""
        self.assertTrue(time_string(time_double(), fmt='%Y-%m-%d') == datetime.now().strftime('%Y-%m-%d'))
        self.assertTrue(time_double('2015-12-15/12:00') == 1450180800.0000000)
        self.assertTrue(time_double('2015-12-15/12') == 1450180800.0000000)
        #self.assertTrue(time_double('2015-12-15/6') == 1450159200.0000000) #this one doesn't work
        self.assertTrue(time_double('2015-12-15/6:00') == 1450159200.0000000)
        self.assertTrue(time_double('2015-12-15/06:00') == 1450159200.0000000)
        self.assertTrue(time_double('2015-12-15') == 1450137600.0000000)
        self.assertTrue(time_double('2015 12 15') == 1450137600.0000000)
        self.assertTrue(time_double('2015-12') == 1448928000.0000000)
        self.assertTrue(time_double('2015') == 1420070400.0000000)
        self.assertTrue(time_double('2015-12-15 12:07:23.767000') == 1450181243.767)
        self.assertTrue(time_double(['2015-12-15 12:07:23.767000', '2015-12-15 12:07:43.767000']) == [1450181243.767, 1450181263.767])

    def test_degap(self):
        float_times=np.array([1.0,2.0,3.0,11.0,12.0,13.0,21.0,22.0,23.0])
        int_times = np.array(float_times,dtype=np.int64)
        str_times = time_string(float_times)
        dtimes=np.array(float_times*1e9,dtype='datetime64[ns]')
        data = [1.0,2.0,3.0,11.0,12.0,13.0,21.0,22.0,23.0]
        sp = np.array([1,2,3])
        spdata = [sp,sp,sp,sp,sp,sp,sp,sp,sp]
        spv = [100,200,300]
        # Scalar data with floating point times
        store_data('degap_test_float',data={'x':float_times,'y':data})
        # Spectral data with floating point times
        store_data('degap_test_float_spec',data={'x':float_times,'y':spdata,'v':spv})
        options('degap_test_float_spec','spec',1)
        # Scalar data with integer times
        store_data('degap_test_int',data={'x':int_times,'y':data})
        # Spectral data with integer times
        store_data('degap_test_int_spec',data={'x':int_times,'y':spdata,'v':spv})
        options('degap_test_int_spec','spec',1)
        # Scalar data with datetime64[ns] times
        store_data('degap_test_dt64',data={'x':dtimes,'y':data})
        # Spectral data with datetime64[ns] times
        store_data('degap_test_dt64_spec',data={'x':dtimes,'y':spdata,'v':spv})
        options('degap_test_dt64_spec','spec',1)
        # Scalar data with string times
        store_data('degap_test_string',data={'x':str_times,'y':data})
        # Spectral data with string times
        store_data('degap_test_string_spec',data={'x':str_times,'y':spdata,'v':spv})
        options('degap_test_string_spec','spec',1)

        degap('degap_test_float',dt=2.0, maxgap=5.0, newname='degap_maxgap')
        dg_data=get_data('degap_maxgap')
        # Should NOT insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],11.0)
        self.assertTrue(np.isfinite(dg_data.y[3]))
        self.assertTrue(len(dg_data.times), 9)

        degap('degap_test_float',dt=2.0, newname='degap_default_gapfill')
        dg_data=get_data('degap_default_gapfill')
        # Should NOT insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertTrue(len(dg_data.times), 15)

        degap('degap_test_float',dt=2.0, onenanpergap=True, newname='degap_onenanpergap')
        dg_data=get_data('degap_onenanpergap')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertEqual(dg_data.times[7],15.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertFalse(np.isfinite(dg_data.y[7]))
        self.assertTrue(len(dg_data.times), 11)

        degap('degap_test_float',dt=2.0, twonanpergap=True, newname='degap_twonanpergap')
        dg_data=get_data('degap_twonanpergap')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],3.25)
        self.assertEqual(dg_data.times[4],10.75)
        self.assertFalse(np.isfinite(dg_data.y[3]))
        self.assertFalse(np.isfinite(dg_data.y[4]))
        self.assertTrue(len(dg_data.times), 13)

        degap('degap_test_float',dt=2.0)
        dg_data=get_data('degap_test_float')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap('degap_test_int',dt=2.0)
        dg_data=get_data('degap_test_int')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap('degap_test_float_spec',dt=2.0)
        dg_data=get_data('degap_test_float_spec')
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap('degap_test_int_spec',dt=2.0)
        dg_data=get_data('degap_test_int_spec')
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap('degap_test_dt64',dt=2.0)
        dg_data=get_data('degap_test_dt64')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap('degap_test_dt64_spec',dt=2.0)
        dg_data=get_data('degap_test_dt64_spec')
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

        degap('degap_test_string',dt=2.0)
        dg_data=get_data('degap_test_string')
        # Should insert a data point with timestamp 5.0, value NaN
        self.assertEqual(dg_data.times[3],5.0)
        self.assertFalse(np.isfinite(dg_data.y[3]))

        degap('degap_test_string_spec',dt=2.0)
        dg_data=get_data('degap_test_string_spec')
        # Should insert a data point with timestamp 5.0, value [NaN, NaN, NaN]
        self.assertEqual(dg_data.times[3], 5.0)
        self.assertFalse(np.isfinite(dg_data.y[3][1]))

    def test_missing_pseudovar(self):
        times=np.array([1.0,2.0,3.0,11.0,12.0,13.0,21.0,22.0,23.0])
        data = [1.0,2.0,3.0,11.0,12.0,13.0,21.0,22.0,23.0]

        store_data('basevar1', data={'x':times,'y':data})
        store_data('basevar2', data={'x':times,'y':data})
        store_data('pseudovar_missing', data=['basevar1', 'basevar2', 'missing_var'])
        self.assertTrue(data_exists('pseudovar_missing'))
        store_data('pseudovar_empty_list',data=[])
        self.assertFalse(data_exists('pseudovar_empty_list'))
        store_data('pseudovar_all_bad',data=['missing_var1','missing_var2','missing_var3'])
        self.assertFalse(data_exists('pseudovar_all_bad'))


if __name__ == '__main__':
    unittest.main()